/************************************************************************
 * EPP RTK C++
 * Copyright (C) 2001, 2002 The Global Name Registry, Limited 
 * 
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact information:  epprtk@gnr.com
 *
 *                       EPP RTK
 *                       The Global Name Registry, Limited
 *                       125 High Holborn
 *                       London WC1V 6QA
 *                       United Kingdom
 ************************************************************************/

#include "npssl.h"
#include <unistd.h>

using namespace std;
using namespace npbase;


char *sslerror(int err) {
   
   static char buf[4196];
   switch (err) {
    case SSL_ERROR_NONE: strcpy(buf, "SSL_ERROR_NONE"); break;
    case SSL_ERROR_ZERO_RETURN: strcpy(buf, "SSL_ERROR_ZERO_RETURN"); break;
    case SSL_ERROR_WANT_READ: strcpy(buf, "SSL_WANT_READ"); break;
    case SSL_ERROR_WANT_WRITE: strcpy(buf, "SSL_WANT_WRITE"); break;
    case SSL_ERROR_WANT_X509_LOOKUP: strcpy(buf, "SSL_WANT_X509_LOOKUP"); break;
    case SSL_ERROR_SYSCALL: strcpy(buf, "SSL_ERROR_SYSCALL"); break;
    case SSL_ERROR_SSL: strcpy(buf, "SSL_ERROR_SSL"); break;
    default: strcpy(buf, "SSL_ERROR_DEFAULT"); break;
   }
   
   return buf;
}


npbase::npssl::npssl(int myBufsiz, const bool mode = true) {
   
   // Create input/output buffer
   bufsiz = myBufsiz;
   buf = new char[(unsigned int)bufsiz];
   
   // Initialize SSL library
   //lint -e534
   // We are not interested in the return values here
   SSL_load_error_strings();
   SSL_library_init();
   //lint +e534
   
   // Add ciphers and message digests
   //lint -e534
   // We are not interested in the return value here
   OpenSSL_add_ssl_algorithms();
   //lint +e534
   
   // create a new SSL_CTX object as 
   // framework for TLS/SSL enabled functions
   ctx = SSL_CTX_new(SSLv23_method());
   
   if (ctx == NULL)
     ssl_error();
   
   server = mode;
   
   ssl = (SSL *)NULL;
}


void npbase::npssl::setClientVerification(int mode,int (*verify_callback)(int, X509_STORE_CTX *)) {
   
   SSL_CTX_set_verify(ctx, mode, verify_callback);
}


void npbase::npssl::setClientVerificationDepth(int depth) {
   
   SSL_CTX_set_verify_depth(ctx, depth);   
}


void npbase::npssl::newConnection(void) {
   
   // create a new SSL structure for a connection  
   ssl = SSL_new(ctx);
   
   if (ssl == NULL)
     ssl_error();
   
   // configure the connection for client mode if necessary  
   if(!server)
     SSL_set_connect_state(ssl);   
}


void npbase::npssl::setExData(int myIndex, void * data) {
   
   // create a new SSL structure for a connection
   //lint -e534
   // We are not interested in the return value here
   SSL_set_ex_data(ssl, myIndex, data);
   //lint +e534
}


X509 * npbase::npssl::getPeerCertificate( void ) {
   
   return SSL_get_peer_certificate(ssl);   
}


long npbase::npssl::getVerifyResult( void) {
   
   return SSL_get_verify_result(ssl);   
} 


npbase::npssl::~npssl() {
   
   // To avoid the destructor from throwing an exception:
   try { 
      SSL_shutdown(ssl);
      SSL_CTX_free(ctx);
      delete[] buf;
   } catch(...) {
   }
   
}


void npbase::npssl::setCertFile(const string &certname) {
   
   if (!SSL_CTX_use_certificate_file(ctx, certname.c_str(), SSL_FILETYPE_PEM))
     ssl_error();
   
   if (!SSL_CTX_use_PrivateKey_file(ctx, certname.c_str(), SSL_FILETYPE_PEM))
     ssl_error();
   
   if (!SSL_CTX_check_private_key(ctx))
     ssl_error();
}


void  npbase::npssl::setVerifyLocations(const string &certname, const string &pathname) {
   
   if(!SSL_CTX_load_verify_locations(ctx, certname.c_str(),pathname.c_str()))
     ssl_error();
}
 

bool npbase::npssl::handShake(int fdin, int fdout) {
   
   // SSL_set_rfd - connect the SSL object with fdin
   if (SSL_set_rfd(ssl, fdin) != 1) {
      // An error occurred 
      ssl_error();
   }
   
   // SSL_set_wfd - connect the SSL object with fdout
   if (SSL_set_wfd(ssl, fdout) != 1) {
      // An error occurred 
      ssl_error();
   }
   
   if(server) {
      if (SSL_accept(ssl) != 1)
	ssl_error();
   } else {
      if (SSL_connect(ssl) != 1)
	ssl_error();
   }
   
   return true;
}


void npbase::npssl::ssl_error(void) {
   
   ERR_print_errors_fp(stderr);

   //lint -e534
   // We are not interested in the return value here
   SSL_shutdown(ssl);
   //lint +e534
   SSL_CTX_free(ctx);
   delete[] buf;
   exit(1);
}


char npbase::npssl::getch(void) {
   
   memset(buf, 0, (unsigned int)bufsiz);
   int bytesread = SSL_read(ssl, buf, 1);
   
   if (bytesread == 0) {
      ERR_print_errors_fp(stderr);
      return EOF;
   }
   
   if (bytesread == -1) {
      ERR_print_errors_fp(stderr);
      return EOF;
   }
   
   return buf[0];
}


const char* npbase::npssl::getchars(long int length_to_read) {
   
   memset(buf, 0, (unsigned int)bufsiz);

   char *temp_ptr = buf;
   while ( length_to_read > 0 ) {

      int bytesread = SSL_read(ssl, temp_ptr, length_to_read);

      if (bytesread == 0) {
         ERR_print_errors_fp(stderr);
         string rep = "0";
         rep[0] = (char)-1;
      }

      if (bytesread == -1) {
         ERR_print_errors_fp(stderr);
         string rep = "0";
         rep[0] = (char)-1;
      }

      length_to_read -= bytesread;
      temp_ptr += bytesread;
   }
   
   return buf;
}


string npbase::npssl::getS(void) {
   
   memset(buf, 0, (unsigned int)bufsiz);
   int bytesread = SSL_read(ssl, buf, bufsiz - 1);
   
   if (bytesread == 0) {
      ERR_print_errors_fp(stderr);
      string rep = "0";
      rep[0] = (char)-1;
      return rep;
   }
   
   if (bytesread == -1) {
      ERR_print_errors_fp(stderr);
      string rep = "0";
      rep[0] = (char)-1;
      return rep;
   }
   
   return buf;
}


string npbase::npssl::bufferedGetS(void) {
   
   if (int_inbuffer.length() == 0) {
      int_inbuffer = getS();
      
      if (int_inbuffer.find(-1) != string::npos) {
	 string s = "0";
	 s[0] = (char) -1;
	 return s;
      }
      
      while (int_inbuffer.find('\n') == string::npos) {
	 int_inbuffer += getS();
	 
	 if (int_inbuffer.find(-1) != string::npos) {
	    string s = "0";
	    s[0] = (char) -1;
	    return s;
	 }
      }
   }
   
   string::size_type pos = int_inbuffer.find('\n');
   string readstr;
   if (pos != string::npos) {
      readstr = int_inbuffer.substr(0, pos + 1);
      int_inbuffer = int_inbuffer.substr(pos + 1);
   } else {
      readstr = int_inbuffer;
      int_inbuffer = "";
   }
   
   return readstr;
}


void npbase::npssl::bufferedUnGetch(char c) {
   
   int_inbuffer = c + int_inbuffer;
   
   return;
}


char npbase::npssl::bufferedGetch(void) {
   
   if (int_inbuffer == "") {
      int_inbuffer = getS();
      
      if (int_inbuffer.find((char)-1) != string::npos)
	return EOF;
      
      while (int_inbuffer.find('\n') == string::npos) {
	 if (int_inbuffer.find((char)-1) != string::npos)
	   return EOF;
	 
	 int_inbuffer += getS();
      }
   }
   
   char c = int_inbuffer[0];
   int_inbuffer = int_inbuffer.substr(1);
   
   return c;
}


bool npbase::npssl::hasMore(void) const {
   
   return int_inbuffer.length() > 0;
}


bool npbase::npssl::putS(const string &str) {
   
   int byteswrote = SSL_write(ssl, str.c_str(), (int)str.length());
   
   if (byteswrote == 0) {
      ERR_print_errors_fp(stderr);
      return false;
   }
   
   if (byteswrote == -1) {
      ERR_print_errors_fp(stderr);
      return false;
   }
   
   return true;
}

bool npbase::npssl::putchars(const char *outdata, int sizeofdata) {
   
   const char *temp_ptr = outdata;
   
   while ( sizeofdata > 0 ) {
      int byteswrote = SSL_write(ssl, temp_ptr, sizeofdata);

      if (byteswrote == 0) {
         ERR_print_errors_fp(stderr);
         return false;
      }

      if (byteswrote == -1) {
         ERR_print_errors_fp(stderr);
         return false;
      }
      sizeofdata -= byteswrote;
      temp_ptr += byteswrote;
   }
   
   return true;
}


bool npbase::npssl::bufferedPutch(char c) {
   
   int_outbuffer += c;
   if (int_outbuffer.find('\n') != string::npos) {
      
      string::size_type pos = int_outbuffer.find('\n');
      string writestr = int_outbuffer.substr(0, pos + 1);
      int_outbuffer = int_outbuffer.substr(pos + 1);
      
      return putS(writestr);
   }
   
   return true;
}


bool npbase::npssl::bufferedPutS(const string &str) {
   int_outbuffer += str;
   
   if (int_outbuffer.find('\n') != string::npos) {
      string::size_type pos = int_outbuffer.find('\n');
      string writestr = int_outbuffer.substr(0, pos + 1);
      int_outbuffer = int_outbuffer.substr(pos + 1);
      
      return putS(writestr);
   }
   
   return true;
}


bool npbase::npssl::flush(void) {
   
   if (int_outbuffer != "") {
      bool res = putS(int_outbuffer);
      int_outbuffer = "";
      return res;
   } else
     return true;
}
